import itertools as it
import sre_constants as sc
import sre_parse
import string
import random as r

def _negate_char_set(char_set):
    return [char for char in string.printable if char not in char_set]

_chars_categories = {
    sc.CATEGORY_DIGIT: string.digits,
    sc.CATEGORY_SPACE: string.whitespace,
    sc.CATEGORY_WORD: string.digits + string.letters + '_',
}
_chars_categories[sc.CATEGORY_NOT_DIGIT] = _negate_char_set(_chars_categories[sc.CATEGORY_DIGIT])
_chars_categories[sc.CATEGORY_NOT_SPACE] = _negate_char_set(_chars_categories[sc.CATEGORY_SPACE])
_chars_categories[sc.CATEGORY_NOT_WORD] = _negate_char_set(_chars_categories[sc.CATEGORY_WORD])

def _generate_branch((something, branches)):
    branch = r.choice(branches)
    return _generate(branch)

def _generate_any(nothing):
    return r.choice(string.printable)

def _generate_in(choices):
    if choices[0][0] == sc.NEGATE:
        chars_to_negate = []
        for element in choices[1:]:
            if element[0] == sc.LITERAL:
                chars_to_negate.append(_generate_literal(element[1]))
            elif element[0] == sc.CATEGORY:
                chars_to_negate += _chars_categories[element[1]]
            else:
                raise Exception("unsupported category %s in negation" % element[0])
        return r.choice(_negate_char_set(chars_to_negate))
    else:
        return _generate_token(*r.choice(choices))

def _generate_category(category):
    return r.choice(_chars_categories[category])

def _generate_range((from_, to)):
    return chr(r.randint(from_, to))

def _generate_not_literal(code):
    return r.choice(_negate_char_set(chr(code)))

def _generate_max_repeat((from_, to, forest)):
    if to > from_ + 20:
        to = from_ + 20
    return ''.join([_generate(forest) for i in xrange(r.randint(from_, to))])

def _generate_subpattern((number, forest)):
    return _generate(forest)

def _generate_token(type_, value):
    generators = {
        sc.LITERAL: chr,
        sc.BRANCH: _generate_branch,
        sc.ANY: _generate_any,
        sc.IN: _generate_in,
        sc.CATEGORY: _generate_category,
        sc.RANGE: _generate_range,
        sc.NOT_LITERAL: _generate_not_literal,
        sc.MAX_REPEAT: _generate_max_repeat,
        sc.SUBPATTERN: _generate_subpattern
    }
    try:
        return generators[type_](value)
    except KeyError, e:
        raise ValueError("Unsupported regular expression construct: %s, %s" % (type_, str(value)))

def _generate(tokens):
    return ''.join([_generate_token(*token) for token in tokens])

def xeger(pattern):
    return _generate(sre_parse.parse(pattern))

